@@ -1,8 +1,51 @@
 # -*- coding: utf-8 -*-
+
+# Copyright OpenStack Foundation and OpenDev Contributors
+#
+# Licensed under the Apache License, Version 2.0 (the "License"); you may
+# not use this file except in compliance with the License.  You may obtain
+# a copy of the License at
+#
+#    http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
+# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
+# License for the specific language governing permissions and limitations
+# under the License.
+
 from __future__ import print_function
 
+import argparse
+import configparser
+import datetime
+import getpass
+import json
+import os
+import re
+import shlex
+import subprocess
+import sys
+import textwrap
+from urllib.parse import urlencode
+from urllib.parse import urljoin
+from urllib.parse import urlparse
+
+import pkg_resources
+import requests
+
+
+VERBOSE = False
+UPDATE = False
+LOCAL_MODE = 'GITREVIEW_LOCAL_MODE' in os.environ
+CONFIGDIR = os.path.expanduser("~/.config/git-review")
+GLOBAL_CONFIG = "/etc/git-review/git-review.conf"
+USER_CONFIG = os.path.join(CONFIGDIR, "git-review.conf")
+DEFAULTS = dict(scheme='ssh', hostname=False, port=None, project=False,
+                branch='master', remote="gerrit", rebase="1",
+                track="0", usepushurl="0", notopic=False, branchauthor="name")
 COPYRIGHT = """\
-Copyright (C) 2011-2012 OpenStack LLC.
+Copyright OpenStack Foundation and OpenDev Contributors
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -16,38 +59,9 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 implied.
 
 See the License for the specific language governing permissions and
-limitations under the License."""
+limitations under the License.
+"""
 
-import argparse
-import datetime
-import getpass
-import json
-import os
-import re
-import shlex
-import six
-import subprocess
-import sys
-import textwrap
-
-import pkg_resources
-import requests
-from six.moves import configparser
-from six.moves import input as do_input
-from six.moves.urllib.parse import urlencode
-from six.moves.urllib.parse import urljoin
-from six.moves.urllib.parse import urlparse
-
-
-VERBOSE = False
-UPDATE = False
-LOCAL_MODE = 'GITREVIEW_LOCAL_MODE' in os.environ
-CONFIGDIR = os.path.expanduser("~/.config/git-review")
-GLOBAL_CONFIG = "/etc/git-review/git-review.conf"
-USER_CONFIG = os.path.join(CONFIGDIR, "git-review.conf")
-DEFAULTS = dict(scheme='ssh', hostname=False, port=None, project=False,
-                branch='master', remote="gerrit", rebase="1",
-                track="0", usepushurl="0")
 
 _branch_name = None
 _has_color = None
@@ -55,6 +69,7 @@ _use_color = None
 _orig_head = None
 _rewrites = None
 _rewrites_push = None
+BRANCHAUTHOR = None
 
 
 class colors(object):
@@ -122,11 +137,7 @@ def run_command_status(*argv, **kwargs):
     if VERBOSE:
         print(datetime.datetime.now(), "Running:", " ".join(argv))
     if len(argv) == 1:
-        # for python2 compatibility with shlex
-        if sys.version_info < (3,) and isinstance(argv[0], six.text_type):
-            argv = shlex.split(argv[0].encode('utf-8'))
-        else:
-            argv = shlex.split(str(argv[0]))
+        argv = shlex.split(str(argv[0]))
     stdin = kwargs.pop('stdin', None)
     newenv = os.environ.copy()
     newenv['LANG'] = 'C'
@@ -161,10 +172,10 @@ def git_credentials(url):
     """Return credentials using git credential or None."""
     cmd = 'git', 'credential', 'fill'
     stdin = 'url=%s' % url
-    rc, out = run_command_status(*cmd, stdin=stdin.encode('utf-8'))
+    rc, out = run_command_status(*cmd, stdin=stdin)
     if rc:
         return None
-    data = dict(l.split('=', 1) for l in out.splitlines())
+    data = dict(line.split('=', 1) for line in out.splitlines())
     return data['username'], data['password']
 
 
@@ -236,7 +247,7 @@ def run_custom_script(action):
     script_file = "%s-review" % (action)
     (top_dir, git_dir) = git_directories()
     paths = [os.path.join(CONFIGDIR, "hooks", script_file),
-             os.path.join(git_dir, "hooks", script_file)]
+             os.path.join(git_get_hooks_path(top_dir, git_dir), script_file)]
     for fpath in paths:
         if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
             status, output = run_command_status(fpath)
@@ -274,7 +285,26 @@ def git_config_get_value(section, option, default=None, as_bool=False):
         raise
 
 
+def git_get_hooks_path(top_dir, git_dir):
+    """Get the path where we need to store and retrieve Git hooks.
+
+    Normally hooks go into .git/hooks, but users can override with the
+    core.hooksPath option. This can either be an absolute path, in
+    which case we use it as-is, or a relative path, in which case we
+    must interpret it as relative to top_dir.
+    """
+    hook_dir = os.path.join(git_dir, "hooks")
+    hooks_path_option = git_config_get_value('core', 'hooksPath')
+    if hooks_path_option:
+        if os.path.isabs(hooks_path_option):
+            hook_dir = hooks_path_option
+        else:
+            hook_dir = os.path.join(top_dir, hooks_path_option)
+    return hook_dir
+
+
 class Config(object):
+
     """Expose as dictionary configuration options."""
 
     def __init__(self, config_file=None):
@@ -352,6 +382,15 @@ def set_hooks_commit_msg(remote, target_file):
                            % (userhost, (":%s" % port) if port else "")
                 print("Fetching commit hook from: %s" % hook_url)
             run_command_exc(CannotInstallHook, *cmd)
+        # If there are submodules, the hook needs to be installed into
+        # each of them.
+        # Here, we don't check for any nonstandard hooks path, because
+        # it should be safe to assume that very few users are inclined
+        # to set the core.hooksPath option in a submodule checkout.
+        run_command_exc(
+            CannotInstallHook,
+            "git", "submodule", "foreach",
+            'cp -p %s "$(git rev-parse --git-dir)/hooks/"' % target_file)
 
     if not os.access(target_file, os.X_OK):
         os.chmod(target_file, os.path.stat.S_IREAD | os.path.stat.S_IEXEC)
@@ -394,7 +433,7 @@ def add_remote(scheme, hostname, port, project, remote, usepushurl):
         print("No remote set, testing %s" % remote_url)
     if not test_remote_url(remote_url):
         print("Could not connect to gerrit.")
-        username = do_input("Enter your gerrit username: ")
+        username = input("Enter your gerrit username: ")
         remote_url = make_remote_url(scheme, username, hostname, port, project)
         print("Trying again with %s" % remote_url)
         if not test_remote_url(remote_url):
@@ -469,8 +508,8 @@ def alias_url(url, rewrite_push):
         # there is a longer insteadOf alias.
         longest = None
         for alias in rewrites:
-            if (url.startswith(alias)
-                    and (longest is None or len(longest) < len(alias))):
+            if url.startswith(alias) and (
+                    longest is None or len(longest) < len(alias)):
                 longest = alias
 
         if longest:
@@ -714,7 +753,9 @@ def load_config_file(config_file):
         'remote': 'defaultremote',
         'rebase': 'defaultrebase',
         'track': 'track',
+        'notopic': 'notopic',
         'usepushurl': 'usepushurl',
+        'branchauthor': 'branchauthor',
     }
     config = {}
     for config_key, option_name in options.items():
@@ -843,6 +884,23 @@ def rebase_changes(branch, remote, interactive=True):
         return False
     _orig_head = output
 
+    # Avoid trying to do a test rebase if there are uncommitted changes.
+    # Either the rebase will fail with a similar message, or if the user
+    # has turned on rebase.autostash then the subsequent reset will
+    # silently discard those changes.
+    cmd = "git diff --ignore-submodules --quiet"
+    (status, output) = run_command_status(cmd)
+    if status != 0:
+        printwrap("You have unstaged changes. Please commit or stash them "
+                  "first, and then try again.")
+        sys.exit(1)
+    cmd = "git diff --cached --ignore-submodules --quiet"
+    (status, output) = run_command_status(cmd)
+    if status != 0:
+        printwrap("You have uncommitted changes. Please commit or stash them "
+                  "first, and then try again.")
+        sys.exit(1)
+
     cmd = "git show-ref --quiet --verify refs/%s" % remote_branch
     (status, output) = run_command_status(cmd)
     if status != 0:
@@ -851,10 +909,25 @@ def rebase_changes(branch, remote, interactive=True):
                   "re-run with the '-R' option enabled." % (branch, remote))
         sys.exit(1)
 
-    if interactive:
-        cmd = "git rebase -p -i %s" % remote_branch
-    else:
-        cmd = "git rebase -p %s" % remote_branch
+    # Determine git version to set rebase flags below.
+    output = run_command("git version")
+    rebase_flag = "--rebase-merges"
+    if "git version" in output:
+        try:
+            v = output.rsplit(None, 1)[1]
+            gitv = tuple(map(int, v.split('.')[:3]))
+            if gitv < (2, 18, 0):
+                rebase_flag = "--preserve-merges"
+        except Exception:
+            # We tried to determine the version and failed. Use current git
+            # flag as fallback.
+            warn("Could not determine git version. "
+                 "Using modern git rebase flags.")
+
+    interactive_flag = interactive and '-i' or ''
+
+    cmd = "git rebase %s %s %s" % \
+        (rebase_flag, interactive_flag, remote_branch)
 
     (status, output) = run_command_status(cmd, GIT_EDITOR='true')
     if status != 0:
@@ -932,7 +1005,7 @@ def assert_one_change(remote, branch, yes, have_hook):
                       "branches (for independent changes).")
             print("\nThe outstanding commits are:\n\n%s\n\n"
                   "Do you really want to submit the above commits?" % output)
-            yes_no = do_input("Type 'yes' to confirm, other to cancel: ")
+            yes_no = input("Type 'yes' to confirm, other to cancel: ")
             if yes_no.lower().strip() != "yes":
                 print("Aborting.")
                 sys.exit(1)
@@ -1113,8 +1186,8 @@ class CheckoutExistingBranchFailed(CommandFailed):
     EXIT_CODE = 65
 
 
-class ResetHardFailed(CommandFailed):
-    "Failed to hard reset downloaded branch"
+class ResetKeepFailed(CommandFailed):
+    "Failed to reset downloaded branch"
     EXIT_CODE = 66
 
 
@@ -1183,7 +1256,8 @@ def fetch_review(review, masterbranch, remote, project):
     except KeyError:
         topic = review
     try:
-        author = re.sub('\W+', '_', review_info['owner']['name']).lower()
+        author = re.sub(r'\W+', '_',
+                        review_info['owner'][BRANCHAUTHOR]).lower()
     except KeyError:
         author = 'unknown'
     remote_branch = review_info['branch']
@@ -1216,7 +1290,7 @@ def checkout_review(branch_name, remote, remote_branch):
                             branch_name)
 
     except CheckoutNewBranchFailed as e:
-        if re.search("already exists\.?", e.output):
+        if re.search(r"already exists\.?", e.output):
             print("Branch %s already exists - reusing" % branch_name)
             track_remote, track_branch = parse_tracking(
                 ref='refs/heads/' + branch_name)
@@ -1229,8 +1303,8 @@ def checkout_review(branch_name, remote, remote_branch):
                 raise BranchTrackingMismatch
             run_command_exc(CheckoutExistingBranchFailed,
                             "git", "checkout", branch_name)
-            run_command_exc(ResetHardFailed,
-                            "git", "reset", "--hard", "FETCH_HEAD")
+            run_command_exc(ResetKeepFailed,
+                            "git", "reset", "--keep", "FETCH_HEAD")
         else:
             raise
 
@@ -1352,21 +1426,21 @@ class _DownloadFlag(argparse.Action):
     def __call__(self, parser, namespace, value, option_string=None):
         url = urlparse(value)
         # Turn URLs into change ids:
-        #   https://review.openstack.org/423436
+        #   https://review.opendev.org/423436
         # and
-        #   https://review.openstack.org/423436/
+        #   https://review.opendev.org/423436/
         # and
-        #   https://review.openstack.org/#/c/423436
+        #   https://review.opendev.org/#/c/423436
         # and
-        #   https://review.openstack.org/c/<project>/+/423436
+        #   https://review.opendev.org/c/<project>/+/423436
         # become
         #   "423436"
         # while
-        #   https://review.openstack.org/423436/1
+        #   https://review.opendev.org/423436/1
         # and
-        #   https://review.openstack.org/#/c/423436/1
+        #   https://review.opendev.org/#/c/423436/1
         # and
-        #   https://review.openstack.org/c/<project>/+/423436/1
+        #   https://review.opendev.org/c/<project>/+/423436/1
         # become
         #   "423436,1".
         #
@@ -1389,7 +1463,26 @@ class _DownloadFlag(argparse.Action):
 def _main():
     usage = "git review [OPTIONS] ... [BRANCH]"
 
-    parser = argparse.ArgumentParser(usage=usage, description=COPYRIGHT)
+    description = """\
+A git command for submitting branches to Gerrit.
+
+git-review is a tool that helps submitting git branches to gerrit for
+review.
+"""
+
+    epilog = """\
+additional information:
+
+  Options configurable using: git config --global gitreview.OPTION VALUE
+
+  branchauthor          name (default), username, email
+"""
+
+    parser = argparse.ArgumentParser(
+        usage=usage,
+        description=description,
+        formatter_class=argparse.RawDescriptionHelpFormatter,
+        epilog=epilog)
 
     topic_arg_group = parser.add_mutually_exclusive_group()
     topic_arg_group.add_argument("-t", "--topic", dest="topic",
@@ -1400,8 +1493,6 @@ def _main():
 
     parser.add_argument("--reviewers", nargs="+",
                         help="Add reviewers to uploaded patch sets.")
-    parser.add_argument("-D", "--draft", dest="draft", action="store_true",
-                        help="Submit review as a draft")
     parser.add_argument("-n", "--dry-run", dest="dry", action="store_true",
                         help="Don't actually submit the branch for review")
     parser.add_argument("-i", "--new-changeid", dest="regenerate",
@@ -1413,6 +1504,10 @@ def _main():
                         action="store_true",
                         help="Use remote push-url logic instead of separate"
                              " remotes")
+    parser.add_argument('--notify',
+                        choices=['NONE', 'OWNER', 'OWNER_REVIEWERS', 'ALL'],
+                        help="Control to whom email notifications are sent,"
+                             " defaults to ALL.")
 
     rebase_group = parser.add_mutually_exclusive_group()
     rebase_group.add_argument("-R", "--no-rebase", dest="rebase",
@@ -1480,6 +1575,9 @@ def _main():
     parser.add_argument("-l", "--list", dest="list", action="count",
                         help="List available reviews for the current project, "
                         "if passed more than once, will show more information")
+    parser.add_argument("--no-thin", dest="no_thin", action="store_true",
+                        help="git push with --no-thin. This may workaround "
+                        "issues with pushing in some circumstances.")
     parser.add_argument("-y", "--yes", dest="yes", action="store_true",
                         help="Indicate that you do, in fact, understand if "
                              "you are submitting more than one patch")
@@ -1530,7 +1628,6 @@ def _main():
     parser.add_argument("branch", nargs="?")
 
     parser.set_defaults(dry=False,
-                        draft=False,
                         verbose=False,
                         update=False,
                         setup=False,
@@ -1551,6 +1648,7 @@ def _main():
         parser.set_defaults(rebase=convert_bool(config['rebase']),
                             track=convert_bool(config['track']),
                             remote=None,
+                            notopic=convert_bool(config['notopic']),
                             usepushurl=convert_bool(config['usepushurl']))
     options = parser.parse_args()
 
@@ -1581,6 +1679,13 @@ def _main():
     yes = options.yes
     status = 0
 
+    global BRANCHAUTHOR
+    BRANCHAUTHOR = config['branchauthor']
+    if BRANCHAUTHOR not in ["name", "email", "username"]:
+        warn("Invalid gitreview.branchauthor: %s (using default: %s)"
+             % (BRANCHAUTHOR, DEFAULTS["branchauthor"]))
+        BRANCHAUTHOR = DEFAULTS["branchauthor"]
+
     if options.track:
         remote, branch = resolve_tracking(remote, branch)
 
@@ -1618,7 +1723,8 @@ def _main():
     if options.custom_script:
         run_custom_script("pre")
 
-    hook_file = os.path.join(git_dir, "hooks", "commit-msg")
+    hook_dir = git_get_hooks_path(top_dir, git_dir)
+    hook_file = os.path.join(hook_dir, "commit-msg")
     have_hook = os.path.exists(hook_file) and os.access(hook_file, os.X_OK)
 
     if not have_hook:
@@ -1636,14 +1742,14 @@ def _main():
             sys.exit(1)
     assert_one_change(remote, branch, yes, have_hook)
 
+    no_thin = ''
+    if options.no_thin:
+        no_thin = '--no-thin'
+
     ref = "for"
 
-    if options.draft:
-        ref = "drafts"
-        if options.custom_script:
-            run_custom_script("draft")
-
-    cmd = "git push %s HEAD:refs/%s/%s" % (remote, ref, branch)
+    cmd = ("git push --no-follow-tags %s %s HEAD:refs/%s/%s" %
+           (no_thin, remote, ref, branch))
     push_options = []
     if options.topic is not None:
         topic = options.topic
@@ -1680,6 +1786,9 @@ def _main():
     if options.remove_private:
         push_options.append('remove-private')
 
+    if options.notify is not None:
+        push_options.append("notify=%s" % options.notify)
+
     if push_options:
         cmd += "%" + ",".join(push_options)
     if options.dry:
@@ -1700,15 +1809,6 @@ def _main():
 
 
 def main():
-    # workaround for avoiding UnicodeEncodeError on print() with older python
-    if sys.version_info[0] < 3:
-        # without reload print would fail even if sys.stdin.encoding
-        # would report utf-8
-        # see: https://stackoverflow.com/a/23847316/99834
-        stdin, stdout, stderr = sys.stdin, sys.stdout, sys.stderr
-        reload(sys)
-        sys.stdin, sys.stdout, sys.stderr = stdin, stdout, stderr
-        sys.setdefaultencoding(os.environ.get('PYTHONIOENCODING', 'utf-8'))
 
     try:
         _main()
